from __future__ import annotations

from typing import TYPE_CHECKING

from platformdirs import user_cache_path

if TYPE_CHECKING:  # pragma: no cover
    from pathlib import Path


def get_user_cache_path() -> Path:
    return user_cache_path(appname="pynyaa", ensure_exists=True).resolve()



from ._compat import IntEnum, StrEnum


class NyaaCategory(StrEnum):
    """Nyaa categories"""

    ANIME = "Anime"
    ANIME_MUSIC_VIDEO = "Anime - Anime Music Video"
    ANIME_ENGLISH_TRANSLATED = "Anime - English-translated"
    ANIME_NON_ENGLISH_TRANSLATED = "Anime - Non-English-translated"
    ANIME_RAW = "Anime - Raw"

    AUDIO = "Audio"
    AUDIO_LOSSLESS = "Audio - Lossless"
    AUDIO_LOSSY = "Audio - Lossy"

    LITERATURE = "Literature"
    LITERATURE_ENGLISH_TRANSLATED = "Literature - English-translated"
    LITERATURE_NON_ENGLISH_TRANSLATED = "Literature - Non-English-translated"
    LITERATURE_RAW = "Literature - Raw"

    LIVE_ACTION = "Live Action"
    LIVE_ACTION_ENGLISH_TRANSLATED = "Live Action - English-translated"
    LIVE_ACTION_IDOL_PROMOTIONAL_VIDEO = "Live Action - Idol/Promotional Video"
    LIVE_ACTION_NON_ENGLISH_TRANSLATED = "Live Action - Non-English-translated"
    LIVE_ACTION_RAW = "Live Action - Raw"

    PICTURES = "Pictures"
    PICTURES_GRAPHICS = "Pictures - Graphics"
    PICTURES_PHOTOS = "Pictures - Photos"

    SOFTWARE = "Software"
    SOFTWARE_APPLICATIONS = "Software - Applications"
    SOFTWARE_GAMES = "Software - Games"

    @property
    def id(self) -> str:
        """
        Returns the ID of the category.

        This ID corresponds to the category as seen in the URL
        `https://nyaa.si/?f=0&c=1_2&q=`, where `c=1_2` is the ID for `Anime - English-translated`.
        """
        mapping = {
            # All, c=0_0
            "All categories": "0_0",
            # Anime, c=1_X
            "Anime": "1_0",
            "Anime - Anime Music Video": "1_1",
            "Anime - English-translated": "1_2",
            "Anime - Non-English-translated": "1_3",
            "Anime - Raw": "1_4",
            # Audio, c=2_X
            "Audio": "2_0",
            "Audio - Lossless": "2_1",
            "Audio - Lossy": "2_2",
            # Literature, c=3_X
            "Literature": "3_0",
            "Literature - English-translated": "3_1",
            "Literature - Non-English-translated": "3_2",
            "Literature - Raw": "3_3",
            # Live Action, c=4_X
            "Live Action": "4_0",
            "Live Action - English-translated": "4_1",
            "Live Action - Idol/Promotional Video": "4_2",
            "Live Action - Non-English-translated": "4_3",
            "Live Action - Raw": "4_4",
            # Pictures, c=5_X
            "Pictures": "5_0",
            "Pictures - Graphics": "5_1",
            "Pictures - Photos": "5_2",
            # Software, c=6_X
            "Software": "6_0",
            "Software - Applications": "6_1",
            "Software - Games": "6_2",
        }

        return mapping.get(self.value, "0_0")


class NyaaFilter(IntEnum):
    """Nyaa search filters"""

    NO_FILTER = 0
    NO_REMAKES = 1
    TRUSTED_ONLY = 2



from pydantic import AnyUrl, Field, UrlConstraints
from typing_extensions import Annotated

SearchLimit = Annotated[int, Field(gt=0, le=75)]
"""Integer with lower and upper limits."""

MagnetUrl = Annotated[AnyUrl, UrlConstraints(allowed_schemes=["magnet"])]
"""Url that only allows magnets."""



from typing_extensions import NamedTuple

from ._compat import metadata


class Version(NamedTuple):
    major: int
    minor: int
    micro: int


def _get_version() -> str:
    """
    Get the version of pynyaa
    """
    try:
        return metadata.version("pynyaa")

    except metadata.PackageNotFoundError:
        return "0.0.0"



from __future__ import annotations

from typing import Any
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from xmltodict import parse as xmltodict_parse


def parse_nyaa_torrent_page(base_url: str, html: str) -> dict[str, Any]:
    """
    Parse the HTML page to extract relevant information

    Parameters
    ----------
    base_url : str
        Base URL used to construct the full URL from relative URLs.
    html : str
        HTML content of the Nyaa page.

    Returns
    -------
    dict[str, Any]
        A dictionary with the relevant information.
    """

    soup = BeautifulSoup(html, "lxml")

    if success := soup.find("div", class_="panel panel-success"):
        # Trusted uploads use `panel-success` for their green color
        body = success
        is_trusted = True
        is_remake = False
    elif remake := soup.find("div", class_="panel panel-danger"):
        # Remake uploads use `panel-danger` for their red color
        body = remake
        is_trusted = False
        is_remake = True
    else:
        # Default uploads use `panel-default`
        body = soup.find("div", class_="panel panel-default")  # type: ignore
        is_trusted = False
        is_remake = False

    title = soup.title.string.replace(":: Nyaa", "").strip()  # type: ignore

    rows = body.find("div", class_="panel-body").find_all("div", class_="row")  # type: ignore

    # ROW ONE
    row_one = rows[0].find_all("div", class_="col-md-5")
    category = row_one[0].get_text().strip()
    date = row_one[1]["data-timestamp"]

    # ROW TWO
    row_two = rows[1].find_all("div", class_="col-md-5")
    submitter = row_two[0].get_text().strip()
    if submitter.lower() != "anonymous":
        submitter_url = urljoin(base_url, row_two[0].find("a").get("href", f"/user/{submitter}"))
        submitter_status = row_two[0].find("a").get("title", None)

        if submitter_status is not None:
            if submitter_status.lower() == "trusted":
                submitter_trusted = True
                submitter_banned = False
            elif submitter_status.lower() == "banned user":
                submitter_trusted = False
                submitter_banned = True
            elif submitter_status.lower() == "banned trusted":
                submitter_trusted = True
                submitter_banned = True
            else:
                submitter_trusted = False
                submitter_banned = False
    else:
        submitter_url = base_url
        submitter_trusted = False
        submitter_banned = False

    seeders = row_two[1].get_text().strip()

    # ROW THREE
    row_three = rows[2].find_all("div", class_="col-md-5")
    information = row_three[0].get_text().strip()
    leechers = row_three[1].get_text().strip()

    # ROW FOUR
    row_four = rows[3].find_all("div", class_="col-md-5")
    completed = row_four[1].get_text().strip()

    # ROW FOOTER
    footer = body.find("div", class_="panel-footer clearfix").find_all("a")  # type: ignore
    torrent_file = urljoin(base_url, footer[0]["href"])
    magnet = footer[1]["href"]

    # DESCRIPTION
    description = soup.find(id="torrent-description").get_text().strip()  # type: ignore

    return dict(
        title=title,
        category=category,
        date=date,
        submitter=dict(
            name=submitter,
            url=submitter_url,
            is_trusted=submitter_trusted,
            is_banned=submitter_banned,
        ),
        information=information,
        seeders=seeders,
        leechers=leechers,
        completed=completed,
        is_trusted=is_trusted,
        is_remake=is_remake,
        description=description,
        torrent_file=torrent_file,
        magnet=magnet,
    )


def parse_nyaa_rss_page(xml: str, limit: int) -> tuple[str, ...]:
    """
    Parse the torrent links out of the RSS page

    Parameters
    ----------
    xml : str
        Nyaa's RSS XML data as a string
    limit : str
        Maximum number of links to parse out of the RSS page

    Returns
    -------
    tuple[str, ...]
        A tuple of torrent links
    """
    try:
        items = xmltodict_parse(xml, encoding="utf-8")["rss"]["channel"]["item"]
    except KeyError:
        return tuple()

    if isinstance(items, dict):  # RSS returns single results as a dict instead of a list
        items = [items]

    if limit > len(items):
        parsed = [item["guid"]["#text"] for item in items]
    else:
        parsed = [item["guid"]["#text"] for item in items[:limit]]

    return tuple(parsed)



"""Compatibility module to for older python versions"""

import sys

if sys.version_info >= (3, 11):
    from enum import IntEnum, StrEnum
else:
    from enum import IntEnum as IntEnumBase

    from strenum import StrEnum

    class IntEnum(IntEnumBase):
        def __str__(self) -> str:
            return str(self.value)


if sys.version_info >= (3, 10):
    from importlib import metadata
else:
    import importlib_metadata as metadata

__all__ = ("metadata", "StrEnum", "IntEnum")



from httpx import HTTPStatusError
from pydantic import ValidationError

from ._clients import AsyncNyaa, Nyaa
from ._enums import NyaaCategory, NyaaFilter
from ._models import NyaaTorrentPage, Submitter
from ._version import Version, _get_version

__version__ = _get_version()
__version_tuple__ = Version(*tuple(map(int, __version__.split("."))))

__all__ = (
    # Clients
    "AsyncNyaa",
    "Nyaa",
    # Enums
    "NyaaCategory",
    "NyaaFilter",
    # Models
    "NyaaTorrentPage",
    "Submitter",
    # Exceptions
    "HTTPStatusError",
    "ValidationError",
)



import argparse

from ._clients import Nyaa


def main() -> None:
    """
    Pretty simple CLI implementation.
    """
    parser = argparse.ArgumentParser(prog="pynyaa", description="Get the JSON representation of a Nyaa torrent page.")
    parser.add_argument("url", type=str, help="Nyaa URL", metavar="https://nyaa.si/view/...")
    parser.add_argument("--indent", "-i", type=int, default=2, help="indentation level")

    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument("--include", "-inc", type=str, nargs="+", default=None, help="keys to include")
    group.add_argument("--exclude", "-exc", type=str, nargs="+", default=None, help="keys to exclude")

    args = parser.parse_args()
    indent = args.indent
    include = args.include
    exclude = args.exclude

    json = Nyaa().get(args.url.strip()).model_dump_json(indent=indent, include=include, exclude=exclude)

    print(json)


if __name__ == "__main__":
    main()



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import Any
from urllib.parse import urljoin

from hishel import CacheClient, FileStorage
from pydantic import validate_call
from torf import Torrent

from .._enums import NyaaCategory, NyaaFilter
from .._models import NyaaTorrentPage
from .._parser import parse_nyaa_rss_page, parse_nyaa_torrent_page
from .._types import SearchLimit
from .._utils import get_user_cache_path


class Nyaa:
    @validate_call
    def __init__(self, base_url: str = "https://nyaa.si/", cache: bool = True, **kwargs: Any) -> None:
        """
        Nyaa client.

        Parameters
        ----------
        base_url : str, optional
            The base URL of Nyaa. Default is `https://nyaa.si/`.
            This is used for constructing the full URL from relative URLs.
        cache : bool, optional
            Whether to enable caching. Default is `True`.
            This will cache the page upon it's first request and then use the cached result
            for any subsequent requests for the same page.
            This helps in avoiding [HTTP 429 Error](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/429) but
            do note some fields like seeders, leechers, and completed are constantly changing and thus caching would
            mean you won't get the latest data on said fields.
        kwargs : Any, optional
            Keyword arguments to pass to the underlying [`httpx.Client`](https://www.python-httpx.org/api/#client)
            used to make the GET request.
        """
        self._base_url = base_url
        self._cache = cache
        self._kwargs = kwargs
        self._extensions = {"force_cache": self._cache, "cache_disabled": not self._cache}
        self._storage = FileStorage(base_path=get_user_cache_path())

    @property
    def base_url(self) -> str:
        """
        This is the base URL, used for constructing the full URL from relative URLs.
        """
        return self._base_url

    @property
    def cache_path(self) -> Path:
        """
        Path where cache files are stored.
        """
        return get_user_cache_path()

    @validate_call
    def get(self, page: int | str) -> NyaaTorrentPage:
        """
        Retrieve information from a Nyaa torrent page.

        Parameters
        ----------
        page : int or str
            The torrent page.
            This can either be a URL like `https://nyaa.si/view/123456`
            or just the ID like `123456`

        Raises
        ------
        ValidationError
            Invalid input
        HTTPStatusError
            Nyaa returned a non 2xx response.

        Returns
        -------
        NyaaTorrentPage
            A NyaaTorrentPage object representing the retrieved data.
        """

        if isinstance(page, int):
            url = urljoin(self._base_url, f"/view/{page}")
            nyaaid = page
        else:
            url = page
            nyaaid = page.split("/")[-1]  # type: ignore

        with CacheClient(storage=self._storage, **self._kwargs) as client:
            nyaa = client.get(url, extensions=self._extensions).raise_for_status()
            parsed = parse_nyaa_torrent_page(self._base_url, nyaa.text)

            # Get the torrent file and convert it to a torf.Torrent object
            torrent_file = client.get(parsed["torrent_file"], extensions=self._extensions).raise_for_status().content
            torrent = Torrent.read_stream(BytesIO(torrent_file))

            return NyaaTorrentPage(id=nyaaid, url=url, torrent=torrent, **parsed)  # type: ignore

    @validate_call
    def search(
        self,
        query: str,
        *,
        category: NyaaCategory | None = None,
        filter: NyaaFilter | None = None,
        limit: SearchLimit = 3,
    ) -> tuple[NyaaTorrentPage, ...]:
        """
        Search for torrents on Nyaa.

        Parameters
        ----------
        query : str
            The search query string.
        category : NyaaCategory, optional
            The category to filter the search. If None, searches all categories.
        filter : NyaaFilter, optional
            The filter to apply to the search results. If None, no filter is applied.
        limit : SearchLimit, optional
            Maximum number of search results to retrieve. Defaults to 3. Maximum is 75.
            Be cautious with this; higher limits increase the number of requests,
            which may trigger rate limiting responses (HTTP 429) or get your IP banned entirely.

        Raises
        ------
        ValidationError
            Invalid input
        HTTPStatusError
            Nyaa returned a non 2xx response.

        Returns
        -------
        tuple[NyaaTorrentPage, ...]
            A tuple of NyaaTorrentPage objects representing the retrieved data.
        """
        with CacheClient(storage=self._storage, **self._kwargs) as client:
            params = dict(
                page="rss",
                f=filter if filter is not None else 0,
                c=category.id if category is not None else "0_0",
                q=query,
            )

            nyaa = client.get(self._base_url, params=params, extensions=self._extensions).raise_for_status()  # type: ignore
            results = parse_nyaa_rss_page(nyaa.text, limit)

            parsed = [self.get(link) for link in results]

            return tuple(parsed)
